import numpy as np
import yaml
from scipy.optimize import fsolve
from scipy.integrate import solve_ivp

from rotation import quatrot, quatmul

class Drone:

    def __init__(self, file="drone.yaml"):
        stream = open(file, "r")
        drone = yaml.load(stream, Loader=yaml.Loader)
        self.NAME = drone['NAME']
        self.RHO = drone['RHO']
        self.G = drone['G']
        self.MASS = drone['MASS']
        self.INERTIA = np.diag(drone["INERTIA"])
        self.P_R_CG = np.array(drone["P_R_CG"])
        self.PROP_KP = drone['PROP_KP']
        self.PROP_KM = drone['PROP_KM']

    def thrust(self, n):
        # prop forces computation / notice that negative thrust is not implemented
        Kp = self.PROP_KP
        T = Kp * n**2
        # prop moments computation
        Km = self.PROP_KM
        N = np.sign(n) * Km * n**2
        return T, N

    def u_int(self, n):
        # position of rotor wrt center of gravity
        P_R_CG = self.P_R_CG
        T, N = self.thrust(n)
        u = np.array([  T[0] + T[1] + T[2] + T[3], 
                        P_R_CG * (T[3] - T[2]), 
                        P_R_CG * (T[0] - T[1]), 
                        N[2] + N[3] - N[0] - N[1]])
        return u

    def dyn(self, x, u):

        G = self.G
        MASS = self.MASS
        INERTIA = self.INERTIA

        # state demultiplexing
        quat = x[3:7]
        vel = x[7:10]
        rot = x[10:13]

        # linear velocity derivative
        dvdt = np.array([0, 0, G]) - 1/MASS * quatrot(quat, np.array([0, 0, u[0]]))

        # angular velocity derivative
        drdt = np.linalg.inv(INERTIA) @ (u[1:4] - np.cross(rot, INERTIA@rot))

        # kinematic equations
        dpdt = vel
        dqdt = 1/2 * quatmul(quat, np.concatenate([[0], rot]))

        return np.concatenate([dpdt, dqdt, dvdt, drdt])

    def trim(self, vh, vz):
        x = np.zeros(13)
        x[3:7] = np.array([1, 0, 0, 0])
        x[7] = vh
        x[9] = vz

        def func(n):
            dx = self.dyn(x, self.u_int(n))
            return dx[9], dx[10], dx[11], dx[12]

        y0 = np.array([100, 100, 100, 100])
        y = fsolve(func, y0)
        return y

    def step(self, x, n, dt):
        func = lambda t, s: self.dyn(s, self.u_int(n))
        sol = solve_ivp(func, (0, dt), x, method='RK45', max_step=0.01)
        return sol.y.T[-1]

    def sim(self, t_span, x0, fctrl):
        func = lambda t, s: self.dyn(s, self.u_int(fctrl(t, s)))
        sol = solve_ivp(func, t_span, x0, method='RK45', max_step=0.01)
        return sol

